@@ -185,6 +185,15 @@ class FizzHandshakeOptions {
     return *this;
   }
 
+  /**
+   * setkeyUpdateThreshold_ configures the server to initiate a key update
+   * after encrypting a certain number of bytes using the same key.
+   */
+  FizzHandshakeOptions& setkeyUpdateThreshold(size_t flag) {
+    keyUpdateThreshold_ = flag;
+    return *this;
+  }
+
   /**
    * `setPreferIoUringSocket` controls whether the accepted client connection
    * should be handled with an io_uring based transport.
@@ -206,10 +215,13 @@ class FizzHandshakeOptions {
       nullptr};
   FizzLoggingCallback* loggingCallback_{nullptr};
   bool handshakeRecordAlignedReads_{false};
+  size_t keyUpdateThreshold_{0};
   bool preferIoUringSocket_{false};
   friend class FizzAcceptorHandshakeHelper;
 };
 
+class SSLContextManager;
+
 class FizzAcceptorHandshakeHelper
     : public wangle::AcceptorHandshakeHelper,
       public fizz::server::AsyncFizzServer::HandshakeCallback,
@@ -218,20 +230,26 @@ class FizzAcceptorHandshakeHelper
  public:
   FizzAcceptorHandshakeHelper(
       std::shared_ptr<const fizz::server::FizzServerContext> context,
+      std::shared_ptr<const SSLContextManager> sslContextManager,
       const folly::SocketAddress& clientAddr,
       std::chrono::steady_clock::time_point acceptTime,
       wangle::TransportInfo& tinfo,
       FizzHandshakeOptions&& options,
       fizz::AsyncFizzBase::TransportOptions transportOptions)
-      : context_(context),
+      : context_(std::move(context)),
+        sslContextManager_(std::move(sslContextManager)),
         tokenBindingContext_(std::move(options.tokenBindingCtx_)),
         clientAddr_(clientAddr),
         acceptTime_(acceptTime),
         tinfo_(tinfo),
         loggingCallback_(options.loggingCallback_),
         handshakeRecordAlignedReads_(options.handshakeRecordAlignedReads_),
+        keyUpdateThreshold_(options.keyUpdateThreshold_),
         preferIoUringSocket_(options.preferIoUringSocket_),
-        transportOptions_(transportOptions) {}
+        transportOptions_(transportOptions) {
+    DCHECK(context_);
+    DCHECK(sslContextManager_);
+  }
 
   void start(
       folly::AsyncSSLSocket::UniquePtr sock,
@@ -270,7 +288,7 @@ class FizzAcceptorHandshakeHelper
       fizz::server::AsyncFizzServer* transport,
       folly::exception_wrapper ex) noexcept override;
   void fizzHandshakeAttemptFallback(
-      std::unique_ptr<folly::IOBuf> clientHello) override;
+      fizz::server::AttemptVersionFallback fallback) override;
 
   // AsyncSSLSocket::HandshakeCallback API
   void handshakeSuc(folly::AsyncSSLSocket* sock) noexcept override;
@@ -284,8 +302,15 @@ class FizzAcceptorHandshakeHelper
       std::unique_ptr<folly::IOBuf> unread) noexcept override;
   void fdDetachFail(const folly::AsyncSocketException& ex) noexcept override;
 
+  /**
+   * Handles SSLContext selection for TLS handshake fallback logic.
+   * Returns the default ssl context if no sni or no context match for sni.
+   */
+  std::shared_ptr<folly::SSLContext> selectSSLCtx(
+      const folly::Optional<std::string>& sni) const;
+
   std::shared_ptr<const fizz::server::FizzServerContext> context_;
-  std::shared_ptr<folly::SSLContext> sslContext_;
+  std::shared_ptr<const SSLContextManager> sslContextManager_;
   std::shared_ptr<fizz::extensions::TokenBindingContext> tokenBindingContext_;
   std::shared_ptr<fizz::extensions::TokenBindingServerExtension>
       tokenBindingExtension_;
@@ -298,8 +323,9 @@ class FizzAcceptorHandshakeHelper
   wangle::SSLErrorEnum sslError_{wangle::SSLErrorEnum::NO_ERROR};
   FizzLoggingCallback* loggingCallback_;
   bool handshakeRecordAlignedReads_{false};
+  size_t keyUpdateThreshold_{0};
 
-  std::unique_ptr<folly::IOBuf> clientHello_;
+  fizz::server::AttemptVersionFallback fallback_;
   bool preferIoUringSocket_{false};
   fizz::AsyncFizzBase::TransportOptions transportOptions_;
 };
@@ -319,6 +345,11 @@ class DefaultToFizzPeekingCallback
     context_ = std::move(context);
   }
 
+  void setSSLContextManager(
+      std::shared_ptr<const SSLContextManager> sslContextManager) {
+    sslContextManager_ = std::move(sslContextManager);
+  }
+
   void setTransportOptions(
       fizz::AsyncFizzBase::TransportOptions transportOptions) {
     transportOptions_ = transportOptions;
@@ -343,9 +374,13 @@ class DefaultToFizzPeekingCallback
       std::chrono::steady_clock::time_point acceptTime,
       wangle::TransportInfo& tinfo) override {
     auto optionsCopy = options_;
+    if (!(context_ && sslContextManager_)) {
+      return nullptr;
+    }
     return wangle::AcceptorHandshakeHelper::UniquePtr(
         new FizzAcceptorHandshakeHelper(
             context_,
+            sslContextManager_,
             clientAddr,
             acceptTime,
             tinfo,
@@ -355,6 +390,7 @@ class DefaultToFizzPeekingCallback
 
  protected:
   std::shared_ptr<const fizz::server::FizzServerContext> context_;
+  std::shared_ptr<const SSLContextManager> sslContextManager_;
   FizzHandshakeOptions options_;
   fizz::AsyncFizzBase::TransportOptions transportOptions_;
 };
