@@ -15,6 +15,7 @@
  */
 
 #include <boost/thread/barrier.hpp>
+
 #include <folly/portability/GMock.h>
 #include <folly/portability/GTest.h>
 #include <wangle/channel/AsyncSocketHandler.h>
@@ -375,7 +376,7 @@ TEST(Pipeline, Concurrent) {
   boost::barrier b{2};
   auto spam = [&] {
     for (int i = 0; i < 100000; i++) {
-      b.wait();
+      b.count_down_and_wait();
       pipeline->read(i);
     }
   };
